/*
 * USB UGCI driver based on basic USB mouse driver.
 * This is designed to work with the flying UGCI, but could be modified to
 * work with driving or fighting easily enough.
 *
 * Adam Wiggins 7/8/99
 *
 */

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/signal.h>
#include <linux/errno.h>
#include <linux/miscdevice.h>
#include <linux/random.h>
#include <linux/poll.h>
#include <linux/init.h>
#include <linux/malloc.h>
#include <linux/module.h>
#include <linux/ugci.h>

#include <asm/spinlock.h>

#include "usb.h"

#define USB_UGCI_MINOR 66

struct ugci_state_internal {
	struct ugci_state state;	/* actual control data */
	int present;			/* this ugci is plugged in */
	int active;			/* someone is has this ugci's device open */
	int ready;			/* the ugci has changed state since the last read */
	wait_queue_head_t wait;		/* for polling */
	struct fasync_struct *fasync;
};

static struct ugci_state_internal static_ugci_state;

spinlock_t usb_ugci_lock = SPIN_LOCK_UNLOCKED;

static int ugci_irq(int state, void *__buffer, int len, void *dev_id)
{
	signed char *data = __buffer;
	struct ugci_state *st;
	unsigned char old_buttons;
	int i;

	/* finding the ugci is easy when there's only one */
	struct ugci_state_internal *ugci = &static_ugci_state; 

	/* if state changes with no one listening, we don't care */
	if (!ugci->active)
		return 1;

	st = &ugci->state;

	/* check report id to determine what kind of update this is */
	switch (data[0])
	{
		case 1:		/* analog joystick */
			st->jx = data[1];
			st->jy = data[2];
			old_buttons = st->buttons;
			st->buttons &= ~(((1 << NUM_GAME_BTNS) - 1) << BTN_1);
			st->buttons |= (data[6] & 0x7) << BTN_1;
			st->buttons |= data[7] << (BTN_1 + 8);
			if (st->buttons != old_buttons)
			{
				for (i = BTN_1; i < BTN_1 + NUM_GAME_BTNS; i++)
					if ((old_buttons & (1 << i)) != (st->buttons & (1 << i)))
					{
						if (old_buttons & (1 << i))
							st->released_count[i]++;
						else
							st->pressed_count[i]++;
					}
			}
			break;
		case 3:		/* coin drop 1 */
			st->coin_count[0] = *((unsigned short *)(data+1));
			st->coin_count[0] |= *((unsigned short *)(data+2)) << 8;
			if (data[3])
			{
				if (!(st->buttons & BTN_START_1_MASK))
				{
					st->buttons |= BTN_START_1_MASK;
					st->pressed_count[BTN_START_1]++;
				}
			}
			else
			{
				if (st->buttons & BTN_START_1_MASK)
				{
					st->buttons &= ~BTN_START_1_MASK;
					st->released_count[BTN_START_1]++;
				}
			}
			break;
		case 4:		/* coin drop 2 */
			st->coin_count[1] = *((unsigned short *)(data+1));
			st->coin_count[1] |= *((unsigned short *)(data+2)) << 8;
			if (data[3])
			{
				if (!(st->buttons & BTN_START_2_MASK))
				{
					st->buttons |= BTN_START_2_MASK;
					st->pressed_count[BTN_START_2]++;
				}
			}
			else
			{
				if (st->buttons & BTN_START_2_MASK)
				{
					st->buttons &= ~BTN_START_2_MASK;
					st->released_count[BTN_START_2]++;
				}
			}
			break;

#if 0
		/* ------------ unimplemented packets follow ------------ */
		case 6:		/* watchdog */
			break;
		case 11:	/* security buffer part 1 */
		case 13:	/* security buffer part 2 */
			break;
		case 14:	/* keyboard emulation */
		case 17:	/* keyboard emulation */
		case 18:	/* keyboard emulation */
		case 19:	/* keyboard emulation */
			break;
#endif
	}

	ugci->ready = 1;

	wake_up_interruptible(&ugci->wait);
	if (ugci->fasync)
		kill_fasync(ugci->fasync, SIGIO);

	return 1;
}

static int ugci_trackball_irq(int state, void *__buffer, int len, void *dev_id)
{
	signed char *data = __buffer;
	struct ugci_state *st;
	unsigned char old_buttons;
	int i;

	/* finding the ugci is easy when there's only one */
	struct ugci_state_internal *ugci = &static_ugci_state; 

	/* if state changes with no one listening, we don't care */
	if (!ugci->active)
		return 1;

	st = &ugci->state;

	if (data[0] != 0 || data[1] != 0)
		printk("trackball: %d, %d\n", data[0], data[1]);

	/* store new trackball data */
	st->tdx += data[0]; /* data[] is signed, so this works */
	st->tdy += data[1];
	old_buttons = st->buttons;
	st->buttons &= ~(((NUM_TB_BTNS << 1) - 1) << BTN_TRACKBALL_1);
	st->buttons |= (data[2] & 0x7) << 13;
	if (old_buttons != st->buttons)
	{
		for (i = BTN_TRACKBALL_1; i < BTN_TRACKBALL_1 + NUM_TB_BTNS; i++)
			if ((old_buttons & (i << 1)) != (st->buttons & (i << 1)))
			{
				if (old_buttons & (i << 1))
					st->released_count[i]++;
				else
					st->pressed_count[i]++;
			}
	}

	ugci->ready = 1;

	wake_up_interruptible(&ugci->wait);
	if (ugci->fasync)
		kill_fasync(ugci->fasync, SIGIO);

	return 1;
}

static int fasync_ugci(int fd, struct file *filp, int on)
{
	int retval;
	struct ugci_state_internal *ugci = &static_ugci_state;

	retval = fasync_helper(fd, filp, on, &ugci->fasync);
	if (retval < 0)
		return retval;
	return 0;
}

static int release_ugci(struct inode * inode, struct file * file)
{
	struct ugci_state_internal *ugci = &static_ugci_state;

	fasync_ugci(-1, file, 0);
	if (--ugci->active)
		return 0;
	return 0;
}

static int open_ugci(struct inode * inode, struct file * file)
{
	struct ugci_state_internal *ugci = &static_ugci_state;

	if (!ugci->present)
		return -EINVAL;
	if (ugci->active++)
		return 0;
	/* flush state */
	memset(&ugci->state, 0, sizeof(struct ugci_state));
	return 0;
}

static ssize_t write_ugci(struct file * file,
       const char * buffer, size_t count, loff_t *ppos)
{
	return -EINVAL;
}

static ssize_t read_ugci(struct file * file, char * buffer, size_t count, loff_t *ppos)
{
	struct ugci_state *st = &static_ugci_state.state;

	if (count != sizeof(struct ugci_state))
		return -1;

	copy_to_user(buffer, st, sizeof(struct ugci_state));

	// Clear out pressed/released counts
	memset(st->pressed_count, 0, NUM_BTNS * sizeof(__u8));
	memset(st->released_count, 0, NUM_BTNS * sizeof(__u8));
	st->tdx = st->tdy = 0;

	return sizeof(struct ugci_state);

//#define EMULATE_PS2
#ifdef EMULATE_PS2
/*
 * Look like a PS/2 mouse, please..
 *
 * The PS/2 protocol is fairly strange, but
 * oh, well, it's at least common..
 */
	int retval = 0;
	static int state = 0;
	struct ugci_state_internal *ugci = &static_ugci_state;
	struct ugci_state *st = &ugci->state;

	if (count) {
		ugci->ready = 0;
		switch (state) {
		case 0: { /* buttons and sign */
			int buttons = st->buttons;
			st->buttons = 0;
			if (st->tdx < 0)
				buttons |= 0x10;
			if (st->tdy < 0)
				buttons |= 0x20;
			put_user(buttons, buffer);
			buffer++;
			retval++;
			state = 1;
			if (!--count)
				break;
		}
		case 1: { /* tdx */
			int tdx = st->tdx;
			st->tdx = 0;
			put_user(tdx, buffer);
			buffer++;
			retval++;
			state = 2;
			if (!--count)
				break;
		}
		case 2:	{ /* tdy */
			int tdy = st->tdy;
			st->tdy = 0;
			put_user(tdy, buffer);
			buffer++;
			retval++;
			state = 0;
			if (!--count)
				break;
		}

		/*
		 * SUBTLE:
		 *
		 * The only way to get here is to do a read() of
		 * more than 3 bytes: if you read a byte at a time
		 * you will just ever see states 0-2, for backwards
		 * compatibility.
		 *
		 * So you can think of this as a packet interface,
		 * where you have arbitrary-sized packets, and you
		 * only ever see the first three bytes when you read
		 * them in small chunks.
		 */
		{ /* fallthrough - dz */
			int dz = 0;
			put_user(dz, buffer);
			buffer++;
			retval++;
			state = 0;
		}
		break;
		}
	}
	return retval;
#endif
}

static unsigned int ugci_poll(struct file *file, poll_table * wait)
{
	struct ugci_state_internal *ugci = &static_ugci_state;

	poll_wait(file, &ugci->wait, wait);
	if (ugci->ready)
		return POLLIN | POLLRDNORM;
	return 0;
}

struct file_operations usb_ugci_fops = {
	NULL,		/* ugci_seek */
	read_ugci,
	write_ugci,
	NULL, 		/* ugci_readdir */
	ugci_poll, 	/* ugci_poll */
	NULL, 		/* ugci_ioctl */
	NULL,		/* ugci_mmap */
	open_ugci,
	NULL,		/* flush */
	release_ugci,
	NULL,
	fasync_ugci,
};

static struct miscdevice usb_ugci = {
	USB_UGCI_MINOR, "USB UGCI", &usb_ugci_fops
};

static int ugci_probe(struct usb_device *dev)
{
	struct usb_interface_descriptor *interface;
	struct usb_endpoint_descriptor *endpoint;
	struct ugci_state_internal *ugci = &static_ugci_state;
	int i;

	if (dev->descriptor.idVendor == 0x078B)
		printk("USB ugci found, product id = 0x%x\n", dev->descriptor.idProduct);
	else
		return -1;

	if (dev->descriptor.bNumConfigurations != 1)
	{
		printk("%d configuration(s) detected, aborting\n", dev->descriptor.bNumConfigurations);
		return -1;
	}

	if (dev->config[0].bNumInterfaces != 2)
	{
		printk("%d interface(s) detected, aborting\n", dev->config[0].bNumInterfaces);
		return -1;
	}

	for (i = 0; i < 2; i++)
	{
		interface = &dev->config[0].altsetting[0].interface[i];

		if (interface->bNumEndpoints != 1)
		{
		    printk("%d endpoint(s) detected, aborting\n", interface->bNumEndpoints);
		    return -1;
		}

		endpoint = &interface->endpoint[0];

		if (!(endpoint->bEndpointAddress & 0x80))
		{
		    printk("Endpoint is outputable, not too sure if that is right, aborting\n");
		    return -1;
		}

		if ((endpoint->bmAttributes & 3) != 3)
		{
		    printk("Endpoint is not interrutable, aborting\n");
		    return -1;
		}

		if (i == 0)
			usb_set_configuration(dev, dev->config[0].bConfigurationValue);

		usb_request_irq(dev, usb_rcvctrlpipe(dev, endpoint->bEndpointAddress), i == 0 ? ugci_irq : ugci_trackball_irq, endpoint->bInterval, NULL);
	}

	ugci->present = 1;

	return 0;
}

static void ugci_disconnect(struct usb_device *dev)
{
	struct ugci_state_internal *ugci = &static_ugci_state;

	/* this might need work */
	ugci->present = 0;
	printk("Mouse disconnected\n");
}

static struct usb_driver ugci_driver = {
	"ugci",
	ugci_probe,
	ugci_disconnect,
	{ NULL, NULL }
};

int usb_ugci_init(void)
{
	struct ugci_state_internal *ugci = &static_ugci_state;

	misc_register(&usb_ugci);

	ugci->present = ugci->active = 0;
	init_waitqueue_head(&ugci->wait);
	ugci->fasync = NULL;

	usb_register(&ugci_driver);
	printk(KERN_INFO "USB HID boot protocol ugci registered.\n");
	return 0;
}

void usb_ugci_cleanup(void)
{
	/* this, too, probably needs work */
	usb_deregister(&ugci_driver);
	misc_deregister(&usb_ugci);
}

#ifdef MODULE
int init_module(void)
{
	return usb_ugci_init();
}

void cleanup_module(void)
{
	usb_ugci_cleanup();
}
#endif

